from unittest import TestCase

from mockfirestore import MockFirestore


class TestMockFirestore(TestCase):
    def test_client_get_all(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2}
        }}
        doc = fs.collection('foo').document('first')
        results = list(fs.get_all([doc]))
        returned_doc_snapshot = results[0].to_dict()
        expected_doc_snapshot = doc.get().to_dict()
        self.assertEqual(returned_doc_snapshot, expected_doc_snapshot)

    def test_client_collections(self):
        fs = MockFirestore()
        fs._data = {
            'foo': {
                'first': {'id': 1},
                'second': {'id': 2}
            },
            'bar': {}
        }
        collections = fs.collections()
        expected_collections = fs._data

        self.assertEqual(len(collections), len(expected_collections))
        for collection in collections:
            self.assertTrue(collection._path[0] in expected_collections)



from typing import Dict, Any, List
from google.cloud import firestore
from datetime import datetime
from contextlib import contextmanager
from unittest import mock

from mockfirestore._helpers import get_document_iterator, get_by_path, set_by_path, delete_by_path

@contextmanager
def mock_firestore_server_timestamp(now: datetime):
    with mock.patch(f'{__name__}.datetime') as mock_mod_datetime:
        mock_mod_datetime.now = mock.Mock(return_value = now)
        yield None

def apply_transformations(document: Dict[str, Any], data: Dict[str, Any]):
    """Handles special fields like INCREMENT."""
    increments = {}
    arr_unions = {}
    arr_deletes = {}
    deletes = []
    timestamps = []

    for key, value in list(get_document_iterator(data)):
        if isinstance(value, firestore.Increment):
            increments[key] = value.value
        elif isinstance(value, firestore.ArrayUnion):
            arr_unions[key] = value.values
        elif isinstance(value, firestore.ArrayRemove):
            arr_deletes[key] = value.values
            delete_by_path(data, key.split('.'))
        elif isinstance(value, type(firestore.DELETE_FIELD)) and str(value) == str(firestore.DELETE_FIELD):
            deletes.append(key)
            delete_by_path(data, key.split('.'))
        elif isinstance(value, type(firestore.SERVER_TIMESTAMP)) and str(value) == str(firestore.SERVER_TIMESTAMP):
            timestamps.append(key)
            delete_by_path(data, key.split('.'))

        # All other transformations can be applied as needed.
        # See #29 for tracking.

    def _update_data(new_values: dict, default: Any):
        for key, value in new_values.items():
            path = key.split('.')

            try:
                item = get_by_path(document, path)
            except (TypeError, KeyError):
                item = default

            set_by_path(data, path, item + value, create_nested=True)

    def _update_data_to(keys: list[str], value: Any):
        for key in keys:
            path = key.split(".")
            set_by_path(data, path, value, create_nested=True)

    _update_data(increments, 0)
    _update_data(arr_unions, [])
    _update_data_to(timestamps, datetime.now().isoformat())

    _apply_updates(document, data)
    _apply_deletes(document, deletes)
    _apply_arr_deletes(document, arr_deletes)


def _apply_updates(document: Dict[str, Any], data: Dict[str, Any]):
    for key, value in data.items():
        path = key.split(".")
        set_by_path(document, path, value, create_nested=True)


def _apply_deletes(document: Dict[str, Any], data: List[str]):
    for key in data:
        path = key.split(".")
        delete_by_path(document, path)


def _apply_arr_deletes(document: Dict[str, Any], data: Dict[str, Any]):
    for key, values_to_delete in data.items():
        path = key.split(".")
        try:
            value = get_by_path(document, path)
        except KeyError:
            continue
        for value_to_delete in values_to_delete:
            try:
                value.remove(value_to_delete)
            except ValueError:
                pass
        set_by_path(document, path, value)



from unittest import TestCase

from google.cloud import firestore

from mockfirestore import MockFirestore, NotFound


class TestDocumentReference(TestCase):

    def test_get_document_by_path(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.document('foo/first').get()
        self.assertEqual({'id': 1}, doc.to_dict())
        self.assertEqual('first', doc.id)

    def test_set_document_by_path(self):
        fs = MockFirestore()
        fs._data = {}
        doc_content = {'id': 'bar'}
        fs.document('foo/doc1/bar/doc2').set(doc_content)
        doc = fs.document('foo/doc1/bar/doc2').get().to_dict()
        self.assertEqual(doc_content, doc)
        
    def test_document_get_returnsDocument(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document('first').get()
        self.assertEqual({'id': 1}, doc.to_dict())
        self.assertEqual('first', doc.id)

    def test_document_get_documentIdEqualsKey(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc_ref = fs.collection('foo').document('first')
        self.assertEqual('first', doc_ref.id)

    def test_document_get_newDocumentReturnsDefaultId(self):
        fs = MockFirestore()
        doc_ref = fs.collection('foo').document()
        doc = doc_ref.get()
        self.assertNotEqual(None, doc_ref.id)
        self.assertFalse(doc.exists)

    def test_document_get_documentDoesNotExist(self):
        fs = MockFirestore()
        fs._data = {'foo': {}}
        doc = fs.collection('foo').document('bar').get().to_dict()
        self.assertEqual({}, doc)

    def test_get_nestedDocument(self):
        fs = MockFirestore()
        fs._data = {'top_collection': {
            'top_document': {
                'id': 1,
                'nested_collection': {
                    'nested_document': {'id': 1.1}
                }
            }
        }}
        doc = fs.collection('top_collection')\
            .document('top_document')\
            .collection('nested_collection')\
            .document('nested_document')\
            .get().to_dict()

        self.assertEqual({'id': 1.1}, doc)

    def test_get_nestedDocument_documentDoesNotExist(self):
        fs = MockFirestore()
        fs._data = {'top_collection': {
            'top_document': {
                'id': 1,
                'nested_collection': {}
            }
        }}
        doc = fs.collection('top_collection')\
            .document('top_document')\
            .collection('nested_collection')\
            .document('nested_document')\
            .get().to_dict()

        self.assertEqual({}, doc)

    def test_document_set_setsContentOfDocument(self):
        fs = MockFirestore()
        fs._data = {'foo': {}}
        doc_content = {'id': 'bar'}
        fs.collection('foo').document('bar').set(doc_content)
        doc = fs.collection('foo').document('bar').get().to_dict()
        self.assertEqual(doc_content, doc)

    def test_document_set_mergeNewValue(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        fs.collection('foo').document('first').set({'updated': True}, merge=True)
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'id': 1, 'updated': True}, doc)

    def test_document_set_mergeNewValueForNonExistentDoc(self):
        fs = MockFirestore()
        fs.collection('foo').document('first').set({'updated': True}, merge=True)
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'updated': True}, doc)

    def test_document_set_overwriteValue(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        fs.collection('foo').document('first').set({'new_id': 1}, merge=False)
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'new_id': 1}, doc)

    def test_document_set_isolation(self):
        fs = MockFirestore()
        fs._data = {'foo': {}}
        doc_content = {'id': 'bar'}
        fs.collection('foo').document('bar').set(doc_content)
        doc_content['id'] = 'new value'
        doc = fs.collection('foo').document('bar').get().to_dict()
        self.assertEqual({'id': 'bar'}, doc)

    def test_document_update_addNewValue(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        fs.collection('foo').document('first').update({'updated': True})
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'id': 1, 'updated': True}, doc)

    def test_document_update_changeExistingValue(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        fs.collection('foo').document('first').update({'id': 2})
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'id': 2}, doc)

    def test_document_update_documentDoesNotExist(self):
        fs = MockFirestore()
        with self.assertRaises(NotFound):
            fs.collection('foo').document('nonexistent').update({'id': 2})
        docsnap = fs.collection('foo').document('nonexistent').get()
        self.assertFalse(docsnap.exists)

    def test_document_update_isolation(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'nested': {'id': 1}}
        }}
        update_doc = {'nested': {'id': 2}}
        fs.collection('foo').document('first').update(update_doc)
        update_doc['nested']['id'] = 3
        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual({'nested': {'id': 2}}, doc)

    def test_document_update_transformerIncrementBasic(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1}
        }}
        fs.collection('foo').document('first').update({'count': firestore.Increment(2)})

        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual(doc, {'count': 3})

    def test_document_update_transformerIncrementNested(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {
                'nested': {'count': 1},
                'other': {'likes': 0},
            }
        }}
        fs.collection('foo').document('first').update({
            'nested': {'count': firestore.Increment(-1)},
            'other': {'likes': firestore.Increment(1), 'smoked': 'salmon'},
        })

        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual(doc, {
            'nested': {'count': 0},
            'other': {'likes': 1, 'smoked': 'salmon'}
        })

    def test_document_update_transformerIncrementNonExistent(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'spicy': 'tuna'}
        }}
        fs.collection('foo').document('first').update({'count': firestore.Increment(1)})

        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual(doc, {'count': 1, 'spicy': 'tuna'})

    def test_document_delete_documentDoesNotExistAfterDelete(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        fs.collection('foo').document('first').delete()
        doc = fs.collection('foo').document('first').get()
        self.assertEqual(False, doc.exists)

    def test_document_parent(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        coll = fs.collection('foo')
        document = coll.document('first')
        self.assertIs(document.parent, coll)

    def test_document_update_transformerArrayUnionBasic(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"arr": [1, 2]}}}
        fs.collection("foo").document("first").update(
            {"arr": firestore.ArrayUnion([3, 4])}
        )
        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc["arr"], [1, 2, 3, 4])

    def test_document_update_transformerArrayUnionNested(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {
                'nested': {'arr': [1]},
                'other': {'labels': ["a"]},
            }
        }}
        fs.collection('foo').document('first').update({
            'nested': {'arr': firestore.ArrayUnion([2])},
            'other': {'labels': firestore.ArrayUnion(["b"]), 'smoked': 'salmon'},
        })

        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual(doc, {
            'nested': {'arr': [1, 2]},
            'other': {'labels': ["a", "b"], 'smoked': 'salmon'}
        })

    def test_document_update_transformerArrayUnionNonExistent(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'spicy': 'tuna'}
        }}
        fs.collection('foo').document('first').update({'arr': firestore.ArrayUnion([1])})

        doc = fs.collection('foo').document('first').get().to_dict()
        self.assertEqual(doc, {'arr': [1], 'spicy': 'tuna'})

    def test_document_update_nestedFieldDotNotation(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"nested": {"value": 1, "unchanged": "foo"}}}}

        fs.collection("foo").document("first").update({"nested.value": 2})

        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc, {"nested": {"value": 2, "unchanged": "foo"}})

    def test_document_update_nestedFieldDotNotationNestedFieldCreation(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"other": None}}}  # non-existent nested field is created

        fs.collection("foo").document("first").update({"nested.value": 2})

        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc, {"nested": {"value": 2}, "other": None})

    def test_document_update_nestedFieldDotNotationMultipleNested(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"other": None}}}

        fs.collection("foo").document("first").update({"nested.subnested.value": 42})

        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc, {"nested": {"subnested": {"value": 42}}, "other": None})

    def test_document_update_nestedFieldDotNotationMultipleNestedWithTransformer(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"other": None}}}

        fs.collection("foo").document("first").update(
            {"nested.subnested.value": firestore.ArrayUnion([1, 3])}
        )

        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc, {"nested": {"subnested": {"value": [1, 3]}}, "other": None})


    def test_document_update_transformerSentinel(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'spicy': 'tuna'}
        }}
        fs.collection('foo').document('first').update({"spicy": firestore.DELETE_FIELD})

        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc, {})

    def test_document_update_transformerArrayRemoveBasic(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"arr": [1, 2, 3, 4]}}}
        fs.collection("foo").document("first").update(
            {"arr": firestore.ArrayRemove([3, 4])}
        )
        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc["arr"], [1, 2])

    def test_document_update_transformerArrayRemoveNonExistentField(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"arr": [1, 2, 3, 4]}}}
        fs.collection("foo").document("first").update(
            {"arr": firestore.ArrayRemove([5])}
        )
        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc["arr"], [1, 2, 3, 4])

    def test_document_update_transformerArrayRemoveNonExistentArray(self):
        fs = MockFirestore()
        fs._data = {"foo": {"first": {"arr": [1, 2, 3, 4]}}}
        fs.collection("foo").document("first").update(
            {"non_existent_array": firestore.ArrayRemove([1, 2])}
        )
        doc = fs.collection("foo").document("first").get().to_dict()
        self.assertEqual(doc["arr"], [1, 2, 3, 4])




from mockfirestore.client import MockFirestore
from typing import AsyncIterator, Any
from google.cloud import firestore
import inspect
from mockfirestore.document import DocumentReference, DocumentSnapshot
from mockfirestore.collection import CollectionReference
from mockfirestore._helpers import get_by_path

class AsyncCollectionReference:
    def __init__(self, collection_reference: CollectionReference):
        _async_proxy_methods(
                source = self,
                target = collection_reference,
                model = firestore.AsyncCollectionReference,
                wrap_name_type_dict = {
                    'document': (AsyncDocumentReference,False),
                    'order_by': (AsyncCollectionReference,False),
                    'limit': (AsyncCollectionReference,False),
                    'where': (AsyncCollectionReference,False),
                    'start_after': (AsyncCollectionReference,False),
                    }
                )

class AsyncDocumentReference:
    def __init__(self, document_reference: DocumentReference):

        _async_proxy_methods(
                source = self,
                target = document_reference,
                model = firestore.AsyncDocumentReference,
                wrap_name_type_dict = {
                    'collection': (AsyncCollectionReference,False),
                    }
                )
        self.document_reference = document_reference

        async def get_wrapper(*args, **kwargs):
            return DocumentSnapshot(self, get_by_path(document_reference._data, document_reference._path))
        setattr(self, 'get', get_wrapper)

class Batch:
    def __init__(self):
        self.ops = []

    def set(self, ref: AsyncDocumentReference, data: dict[str, Any]):
        self.ops.append(('set', ref, data))

    async def commit(self):
        for op_meth, op_p1, op_p2 in self.ops:
            if op_meth == 'set':
                await op_p1.set(op_p2)

class AsyncMockFirestore(MockFirestore):
    def collection(self, path: str) -> AsyncCollectionReference:
        return AsyncCollectionReference(super().collection(path))

    def batch(self):
        return Batch()

def _async_proxy_methods(
        source,
        target,
        model,
        wrap_name_type_dict: dict[str, tuple[type, bool]] | None = None):
    wrap_name_type_dict = wrap_name_type_dict or {}
    for member_name, member_type in inspect.getmembers(model):
        if wrap_type_is_async := wrap_name_type_dict.get(member_name):
            _async_wrap_method(source, target, member_name, wrap_type_is_async[0], wrap_type_is_async[1])
            continue
        target_method = getattr(target, member_name, None)
        if not target_method:
            continue
        if proxy_method := _get_proxy_method(target_method, member_type):
            setattr(source, member_name, proxy_method)

def _get_proxy_method(target_method, model_member_type):
    if inspect.isasyncgenfunction(model_member_type):
        async def proxy_method(*args, **kwargs):
            for d in target_method(*args, **kwargs):
                yield d
        return proxy_method
    if inspect.iscoroutinefunction(model_member_type):
        async def proxy_method(*args, **kwargs):
            return target_method(*args, **kwargs)
        return proxy_method
    if inspect.isfunction(model_member_type):
        return target_method
    if isinstance(model_member_type, property):
        return target_method

def _async_wrap_method(you, target, method_name: str, wrap_type: type, is_async: bool):
    if is_async:
        async def wrapper(*args, **kwargs):
            return wrap_type(getattr(target, method_name)(*args, **kwargs))
    else:
        def wrapper(*args, **kwargs):
            return wrap_type(getattr(target, method_name)(*args, **kwargs))
    setattr(you, method_name, wrapper)



from unittest import TestCase

from mockfirestore import MockFirestore, DocumentReference, DocumentSnapshot, AlreadyExists


class TestCollectionReference(TestCase):
    def test_collection_get_returnsDocuments(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2}
        }}
        docs = list(fs.collection('foo').stream())

        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual({'id': 2}, docs[1].to_dict())

    def test_collection_get_collectionDoesNotExist(self):
        fs = MockFirestore()
        docs = fs.collection('foo').stream()
        self.assertEqual([], list(docs))

    def test_collection_get_nestedCollection(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {
                'id': 1,
                'bar': {
                    'first_nested': {'id': 1.1}
                }
            }
        }}
        docs = list(fs.collection('foo').document('first').collection('bar').stream())
        self.assertEqual({'id': 1.1}, docs[0].to_dict())

    def test_collection_get_nestedCollection_by_path(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {
                'id': 1,
                'bar': {
                    'first_nested': {'id': 1.1}
                }
            }
        }}
        docs = list(fs.collection('foo/first/bar').stream())
        self.assertEqual({'id': 1.1}, docs[0].to_dict())

    def test_collection_get_nestedCollection_collectionDoesNotExist(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        docs = list(fs.collection('foo').document('first').collection('bar').stream())
        self.assertEqual([], docs)

    def test_collection_get_nestedCollection_by_path_collectionDoesNotExist(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        docs = list(fs.collection('foo/first/bar').stream())
        self.assertEqual([], docs)

    def test_collection_get_ordersByAscendingDocumentId_byDefault(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'beta': {'id': 1},
            'alpha': {'id': 2}
        }}
        docs = list(fs.collection('foo').stream())
        self.assertEqual({'id': 2}, docs[0].to_dict())

    def test_collection_whereEquals(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'valid': True},
            'second': {'gumby': False}
        }}

        docs = list(fs.collection('foo').where('valid', '==', True).stream())
        self.assertEqual({'valid': True}, docs[0].to_dict())

    def test_collection_whereNotEquals(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('count', '!=', 1).stream())
        self.assertEqual({'count': 5}, docs[0].to_dict())

    def test_collection_whereLessThan(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('count', '<', 5).stream())
        self.assertEqual({'count': 1}, docs[0].to_dict())

    def test_collection_whereLessThanOrEqual(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('count', '<=', 5).stream())
        self.assertEqual({'count': 1}, docs[0].to_dict())
        self.assertEqual({'count': 5}, docs[1].to_dict())

    def test_collection_whereGreaterThan(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('count', '>', 1).stream())
        self.assertEqual({'count': 5}, docs[0].to_dict())

    def test_collection_whereGreaterThanOrEqual(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('count', '>=', 1).stream())
        self.assertEqual({'count': 1}, docs[0].to_dict())
        self.assertEqual({'count': 5}, docs[1].to_dict())

    def test_collection_whereMissingField(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'count': 1},
            'second': {'count': 5}
        }}

        docs = list(fs.collection('foo').where('no_field', '==', 1).stream())
        self.assertEqual(len(docs), 0)

    def test_collection_whereNestedField(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'nested': {'a': 1}},
            'second': {'nested': {'a': 2}}
        }}

        docs = list(fs.collection('foo').where('nested.a', '==', 1).stream())
        self.assertEqual(len(docs), 1)
        self.assertEqual({'nested': {'a': 1}}, docs[0].to_dict())

    def test_collection_whereIn(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'field': 'a1'},
            'second': {'field': 'a2'},
            'third': {'field': 'a3'},
            'fourth': {'field': 'a4'},
        }}

        docs = list(fs.collection('foo').where('field', 'in', ['a1', 'a3']).stream())
        self.assertEqual(len(docs), 2)
        self.assertEqual({'field': 'a1'}, docs[0].to_dict())
        self.assertEqual({'field': 'a3'}, docs[1].to_dict())

    def test_collection_whereArrayContains(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'field': ['val4']},
            'second': {'field': ['val3', 'val2']},
            'third': {'field': ['val3', 'val2', 'val1']}
        }}

        docs = list(fs.collection('foo').where('field', 'array_contains', 'val1').stream())
        self.assertEqual(len(docs), 1)
        self.assertEqual(docs[0].to_dict(), {'field': ['val3', 'val2', 'val1']})

    def test_collection_whereArrayContainsAny(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'field': ['val4']},
            'second': {'field': ['val3', 'val2']},
            'third': {'field': ['val3', 'val2', 'val1']}
        }}

        contains_any_docs = list(fs.collection('foo').where('field', 'array_contains_any', ['val1', 'val4']).stream())
        self.assertEqual(len(contains_any_docs), 2)
        self.assertEqual({'field': ['val4']}, contains_any_docs[0].to_dict())
        self.assertEqual({'field': ['val3', 'val2', 'val1']}, contains_any_docs[1].to_dict())

    def test_collection_orderBy(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 1}
        }}

        docs = list(fs.collection('foo').order_by('order').stream())
        self.assertEqual({'order': 1}, docs[0].to_dict())
        self.assertEqual({'order': 2}, docs[1].to_dict())

    def test_collection_orderBy_descending(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 3},
            'third': {'order': 1}
        }}

        docs = list(fs.collection('foo').order_by('order', direction="DESCENDING").stream())
        self.assertEqual({'order': 3}, docs[0].to_dict())
        self.assertEqual({'order': 2}, docs[1].to_dict())
        self.assertEqual({'order': 1}, docs[2].to_dict())

    def test_collection_limit(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2}
        }}
        docs = list(fs.collection('foo').limit(1).stream())
        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual(1, len(docs))

    def test_collection_offset(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').offset(1).stream())

        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual({'id': 3}, docs[1].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_orderby_offset(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').order_by("id").offset(1).stream())

        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual({'id': 3}, docs[1].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_start_at(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').start_at({'id': 2}).stream())
        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_start_at_order_by(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').order_by('id').start_at({'id': 2}).stream())
        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_start_at_doc_snapshot(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3},
            'fourth': {'id': 4},
            'fifth': {'id': 5},
        }}

        doc = fs.collection('foo').document('second').get()

        docs = list(fs.collection('foo').order_by('id').start_at(doc).stream())
        self.assertEqual(4, len(docs))
        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual({'id': 3}, docs[1].to_dict())
        self.assertEqual({'id': 4}, docs[2].to_dict())
        self.assertEqual({'id': 5}, docs[3].to_dict())

    def test_collection_start_after(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').start_after({'id': 1}).stream())
        self.assertEqual({'id': 2}, docs[0].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_start_after_similar_objects(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1, 'value': 1},
            'second': {'id': 2, 'value': 2},
            'third': {'id': 3, 'value': 2},
            'fourth': {'id': 4, 'value': 3}
        }}
        docs = list(fs.collection('foo').order_by('id').start_after({'id': 3, 'value': 2}).stream())
        self.assertEqual({'id': 4, 'value': 3}, docs[0].to_dict())
        self.assertEqual(1, len(docs))

    def test_collection_start_after_order_by(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').order_by('id').start_after({'id': 2}).stream())
        self.assertEqual({'id': 3}, docs[0].to_dict())
        self.assertEqual(1, len(docs))

    def test_collection_start_after_doc_snapshot(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'second': {'id': 2},
            'third': {'id': 3},
            'fourth': {'id': 4},
            'fifth': {'id': 5},
        }}

        doc = fs.collection('foo').document('second').get()

        docs = list(fs.collection('foo').order_by('id').start_after(doc).stream())
        self.assertEqual(3, len(docs))
        self.assertEqual({'id': 3}, docs[0].to_dict())
        self.assertEqual({'id': 4}, docs[1].to_dict())
        self.assertEqual({'id': 5}, docs[2].to_dict())

    def test_collection_end_before(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').end_before({'id': 2}).stream())
        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual(1, len(docs))

    def test_collection_end_before_order_by(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').order_by('id').end_before({'id': 2}).stream())
        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual(1, len(docs))

    def test_collection_end_before_doc_snapshot(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3},
            'fourth': {'id': 4},
            'fifth': {'id': 5},
        }}

        doc = fs.collection('foo').document('fourth').get()

        docs = list(fs.collection('foo').order_by('id').end_before(doc).stream())
        self.assertEqual(3, len(docs))

        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual({'id': 2}, docs[1].to_dict())
        self.assertEqual({'id': 3}, docs[2].to_dict())

    def test_collection_end_at(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').end_at({'id': 2}).stream())
        self.assertEqual({'id': 2}, docs[1].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_end_at_order_by(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3}
        }}
        docs = list(fs.collection('foo').order_by('id').end_at({'id': 2}).stream())
        self.assertEqual({'id': 2}, docs[1].to_dict())
        self.assertEqual(2, len(docs))

    def test_collection_end_at_doc_snapshot(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1},
            'second': {'id': 2},
            'third': {'id': 3},
            'fourth': {'id': 4},
            'fifth': {'id': 5},
        }}

        doc = fs.collection('foo').document('fourth').get()

        docs = list(fs.collection('foo').order_by('id').end_at(doc).stream())
        self.assertEqual(4, len(docs))

        self.assertEqual({'id': 1}, docs[0].to_dict())
        self.assertEqual({'id': 2}, docs[1].to_dict())
        self.assertEqual({'id': 3}, docs[2].to_dict())
        self.assertEqual({'id': 4}, docs[3].to_dict())

    def test_collection_limitAndOrderBy(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 1},
            'third': {'order': 3}
        }}
        docs = list(fs.collection('foo').order_by('order').limit(2).stream())
        self.assertEqual({'order': 1}, docs[0].to_dict())
        self.assertEqual({'order': 2}, docs[1].to_dict())

    def test_collection_listDocuments(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 1},
            'third': {'order': 3}
        }}
        doc_refs = list(fs.collection('foo').list_documents())
        self.assertEqual(3, len(doc_refs))
        for doc_ref in doc_refs:
            self.assertIsInstance(doc_ref, DocumentReference)

    def test_collection_stream(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 1},
            'third': {'order': 3}
        }}
        doc_snapshots = list(fs.collection('foo').stream())
        self.assertEqual(3, len(doc_snapshots))
        for doc_snapshot in doc_snapshots:
            self.assertIsInstance(doc_snapshot, DocumentSnapshot)

    def test_collection_parent(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'order': 2},
            'second': {'order': 1},
            'third': {'order': 3}
        }}
        doc_snapshots = fs.collection('foo').stream()
        for doc_snapshot in doc_snapshots:
            doc_reference = doc_snapshot.reference
            subcollection = doc_reference.collection('order')
            self.assertIs(subcollection.parent, doc_reference)

    def test_collection_addDocument(self):
        fs = MockFirestore()
        fs._data = {'foo': {}}
        doc_id = 'bar'
        doc_content = {'id': doc_id, 'xy': 'z'}
        timestamp, doc_ref = fs.collection('foo').add(doc_content)
        self.assertEqual(doc_content, doc_ref.get().to_dict())

        doc = fs.collection('foo').document(doc_id).get().to_dict()
        self.assertEqual(doc_content, doc)

        with self.assertRaises(AlreadyExists):
            fs.collection('foo').add(doc_content)

    def test_collection_useDocumentIdKwarg(self):
        fs = MockFirestore()
        fs._data = {'foo': {
            'first': {'id': 1}
        }}
        doc = fs.collection('foo').document(document_id='first').get()
        self.assertEqual({'id': 1}, doc.to_dict())



from typing import Iterable, Sequence
from mockfirestore.collection import CollectionReference
from mockfirestore.document import DocumentReference, DocumentSnapshot
from mockfirestore.transaction import Transaction


class MockFirestore:

    def __init__(self) -> None:
        self._data = {}

    def _ensure_path(self, path):
        current_position = self

        for el in path[:-1]:
            if type(current_position) in (MockFirestore, DocumentReference):
                current_position = current_position.collection(el)
            else:
                current_position = current_position.document(el)

        return current_position

    def document(self, path: str) -> DocumentReference:
        path = path.split("/")

        if len(path) % 2 != 0:
            raise Exception("Cannot create document at path {}".format(path))
        current_position = self._ensure_path(path)

        return current_position.document(path[-1])

    def collection(self, path: str) -> CollectionReference:
        path = path.split("/")

        if len(path) % 2 != 1:
            raise Exception("Cannot create collection at path {}".format(path))

        name = path[-1]
        if len(path) > 1:
            current_position = self._ensure_path(path)
            return current_position.collection(name)
        else:
            if name not in self._data:
                self._data[name] = {}
            return CollectionReference(self._data, [name])

    def collections(self) -> Sequence[CollectionReference]:
        return [CollectionReference(self._data, [collection_name]) for collection_name in self._data]

    def reset(self):
        self._data = {}

    def get_all(self, references: Iterable[DocumentReference],
                field_paths=None,
                transaction=None) -> Iterable[DocumentSnapshot]:
        for doc_ref in set(references):
            yield doc_ref.get()

    def transaction(self, **kwargs) -> Transaction:
        return Transaction(self, **kwargs)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="firestore-mock",
    version="0.1.4",
    author="Peter Metcalf",
    description="In-memory implementation of Google Cloud Firestore for use in tests with support for asyncio",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pckmsolutions/firestore-mock",
    packages=setuptools.find_packages(),
    test_suite='',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        "License :: OSI Approved :: MIT License",
    ],
)
